/*
 *  Copyright (c) 2023 The WebM project authors. All Rights Reserved.
 *
 *  Use of this source code is governed by a BSD-style license
 *  that can be found in the LICENSE file in the root of the source
 *  tree. An additional intellectual property rights grant can be found
 *  in the file PATENTS.  All contributing project authors may
 *  be found in the AUTHORS file in the root of the source tree.
 */

#ifndef TEST_INIT_VPX_TEST_H_
#define TEST_INIT_VPX_TEST_H_

namespace libvpx_test {
void init_vpx_test();
}

#endif  // TEST_INIT_VPX_TEST_H_
